"""Miscellaneous functions and classes."""

from json import dumps, loads
from typing import Dict, Union

from monzo.handlers.storage import Storage


class FileSystem(Storage):
    """Class that will store credentials on the file system."""

    __slots__ = [
        '_access_token',
        '_client_id',
        '_client_secret',
        '_expiry',
        '_file',
        '_refresh_token',
        '_fetched',
    ]

    def __init__(self, file: str):
        """
        Initialize FileSystem.

        Args:
            file: THe full path (including filename) to the storage file
        """
        self._file = file

        self._access_token: str = ''
        self._client_id: str = ''
        self._client_secret: str = ''
        self._expiry: int = 0
        self._refresh_token: str = ''
        self._fetched: bool = False

    def set_client_details(self, client_id: str, client_secret: str):
        """
        Store a given client id and secret and clear all other details.

        Args:
            client_id: Client ID
            client_secret: Client secret
        """
        self._access_token = ''
        self._client_id = client_id
        self._client_secret = client_secret
        self._expiry = 0
        self._refresh_token = ''
        self._fetched = False

        content = {
            'access_token': self._access_token,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'expiry': self._expiry,
            'refresh_token': self._refresh_token
        }
        with open(self._file, 'w') as handler:
            handler.write(dumps(content))

    def store(
        self,
        access_token: str,
        client_id: str,
        client_secret: str,
        expiry: int,
        refresh_token: str = ''
    ) -> None:
        """
        Store the Monzo credentials.

        Args:
            access_token: New access token
            client_id: Monzo client ID
            client_secret: Monzo client secret
            expiry: Access token expiry as a unix timestamp
            refresh_token: Refresh token that can be used to renew an access token
        """
        self._fetched = True
        self._access_token = access_token
        self._client_id = client_id
        self._client_secret = client_secret
        self._expiry = expiry
        self._refresh_token = refresh_token

        content = {
            'access_token': access_token,
            'client_id': client_id,
            'client_secret': client_secret,
            'expiry': expiry,
            'refresh_token': refresh_token
        }
        with open(self._file, 'w') as handler:
            handler.write(dumps(content))

    def fetch(self) -> Dict[str, Union[int, str]]:
        """
        Fetch the Monzo credentials previously stored.

        Returns:
            Dictionary containing access token, expiry and refresh token
        """
        try:
            with open(self._file, 'r') as handler:
                content = loads(handler.read())
                self._fetched = True
        except FileNotFoundError:
            content = {}

        for item, value in content.items():
            self.__setattr__(f'_{item}', value)
        return content

    @property
    def is_configured(self) -> bool:
        """
        Identify if Monzo has been configured.

        Returns:
             True if configured otherwise False
        """
        if not self._fetched:
            self.fetch()
        if not self._fetched:
            return False
        keys = [
            'client_id',
            'client_secret',
            'access_token',
            'expiry',
        ]
        print(all(self.__getattribute__(key) for key in keys))
        return all(self.__getattribute__(key) for key in keys)

    @property
    def access_token(self) -> str:
        """
        Property for access token.

        Returns:
             access token
        """
        if not self._fetched:
            self.fetch()
        return self._access_token or ''

    @property
    def client_id(self) -> str:
        """
        Property for client id.

        Returns:
            client id
        """
        if not self._fetched:
            self.fetch()
        return self._client_id or ''

    @client_id.setter
    def client_id(self, client_id: str):
        """
        Setter for client id.

        Args:
            client_id: Monzo client id
        """
        self._client_id = client_id

    @property
    def client_secret(self) -> str:
        """
        Property for client secret.

        Returns:
            client secret
        """
        if not self._fetched:
            self.fetch()
        return self._client_secret or ''

    @client_secret.setter
    def client_secret(self, client_secret: str):
        """
        Setter for client secret.

        Args:
            client_secret: Monzo client secret
        """
        self._client_secret = client_secret

    @property
    def expiry(self) -> int:
        """
        Property for expiry.

        Returns:
            expiry
        """
        if not self._fetched:
            self.fetch()
        return self._expiry or 0

    @property
    def refresh_token(self) -> str:
        """
        Property for refresh token.

        Returns:
            refresh token
        """
        if not self._fetched:
            self.fetch()
        return self._refresh_token or ''



"""Entry point for the viewer."""
from datetime import datetime, timedelta
from json import dumps, loads
from typing import Any, Dict, Optional

from flask import Flask, escape, redirect, render_template, request
from monzo.authentication import Authentication
from monzo.endpoints.account import Account
from monzo.endpoints.transaction import Transaction
from monzo.exceptions import MonzoAuthenticationError, MonzoPermissionsError, MonzoServerError

from monzo_viewer.misc import FileSystem

app = Flask(__name__, template_folder='templates')

REDIRECT_URL = 'http://127.0.0.1:5000/setup/callback'
MONZO_HANDLER = FileSystem(file='monzo.json')


def auth_setup(client_id: str = None, client_secret: str = None):
    """
    Create and set up the Auth object.

    Args:
        client_id: Monzo Client ID, if not specified it will use the value from the handler.
        client_secret: Monzo Client Secret, if not specified it will use the value from the handler.

    Returns:
        Auth object with handler specified
    """
    if not client_id:
        client_id = MONZO_HANDLER.client_id
    if not client_secret:
        client_secret = MONZO_HANDLER.client_secret
    auth = Authentication(
        access_token=MONZO_HANDLER.access_token,
        access_token_expiry=MONZO_HANDLER.expiry,
        client_id=client_id,
        client_secret=client_secret,
        redirect_url=REDIRECT_URL,
        refresh_token=MONZO_HANDLER.refresh_token,
    )
    auth.register_callback_handler(MONZO_HANDLER)
    return auth


@app.route('/setup/', methods=['GET', 'POST'])
def setup():
    """Handle the initial Monzo setup."""
    context = {
        'REDIRECT_URL': REDIRECT_URL
    }
    if request.method == 'GET':
        return render_template('setup.html', **context)
    context['CLIENT_ID_VALUE'] = escape(request.form['client_id'])
    context['CLIENT_SECRET_VALUE'] = escape(request.form['client_secret'])
    if not len(context['CLIENT_ID_VALUE']) or not len(context['CLIENT_SECRET_VALUE']):
        context['ERROR_MESSAGE'] = 'Ensure you enter both a Client ID and a Client Secret'
        return render_template('setup.html', **context)
    else:
        MONZO_HANDLER.set_client_details(
            client_id=context['CLIENT_ID_VALUE'],
            client_secret=context['CLIENT_SECRET_VALUE'],
        )
        auth = auth_setup(client_id=MONZO_HANDLER.client_id, client_secret=MONZO_HANDLER.client_secret)
        return redirect(auth.authentication_url)


@app.route('/setup/callback/')
def setup_callback():
    """Handle the callback requests from Monzo."""
    context = {}
    if all(["code" in request.args, "state" in request.args]):
        code = request.args["code"]
        state = request.args["state"]
        auth = Authentication(
            client_id=MONZO_HANDLER.client_id,
            client_secret=MONZO_HANDLER.client_secret,
            redirect_url=REDIRECT_URL,
        )
        auth.register_callback_handler(handler=MONZO_HANDLER)
        try:
            auth.authenticate(authorization_token=code, state_token=state)
            return render_template('success.html')
        except MonzoAuthenticationError:
            context['error'] = 'Monzo authentication error'
        except MonzoServerError:
            context['error'] = 'Monzo server error'
    else:
        context['error'] = 'Missing parameters'
    return render_template('error.html', **context)


@app.route('/', methods=['GET', 'POST'])
def index():
    """Handle displaying accounts."""
    if not MONZO_HANDLER.is_configured:
        return redirect('/setup/')
    auth = auth_setup()
    if request.method == 'POST' and escape(request.form['account']):
        return redirect(f"/accounts/{escape(request.form['account'])}/")
    context = {'accounts': Account.fetch(auth=auth)}
    return render_template('index.html', **context)


@app.route('/accounts/<account>/')
def transactions_for_account(account: str):
    """Handle displaying transactions for an account."""
    auth = auth_setup()
    thirty_days_ago = datetime.now() - timedelta(days=30)
    context = {
        'transactions': Transaction.fetch(
            auth=auth,
            account_id=account,
            since=thirty_days_ago,
            expand=['merchant'],
        )
    }

    return render_template('transactions.html', **context)


@app.route('/raw_request/', methods=['GET', 'POST'])
def raw_request():
    """Handle the raw_request url."""
    context = {}
    auth = auth_setup()

    if request.method == 'GET':
        return render_template('raw_request.html', **context)

    if 'submit' in request.form:
        authenticated = bool(request.form.get('authenticated', True))
        request_type = request.form.get('authenticated', ['get'])
        headers = loads(request.form['headers']) if 'headers' in request.form else {}
        monzo_parameters = {}
        if 'parameters' in request.form and request.form['parameters']:
            monzo_parameters = loads(request.form['parameters'])

        records = get_raw_request(
            auth=auth,
            path=request.form.get('path', ['/']),
            authenticated=authenticated,
            request_type=request_type,
            headers=headers,
            parameters=monzo_parameters
        )
        context['records'] = dumps(records, indent=4, sort_keys=True)
    return render_template('raw_request.html', **context)


def get_raw_request(
        auth: Authentication,
        path: str,
        authenticated: bool = True,
        request_type: str = 'get',
        headers: Optional[Dict[str, Any]] = None,
        parameters: Optional[Dict[str, Any]] = None,
):
    """
    Perform a raw request.

    Args
        auth: Monzo Authentication object
        path: API path for request
        authenticated: True if call should be authenticated
        request_type: HTTP method to use (DELETE, GET, POST, PUT)
        headers: Dictionary of headers for the request
        parameters: Dictionary of parameters for the request

    Returns:
        List of Account objects
    """
    try:
        res = auth.make_request(
            path=path,
            authenticated=authenticated,
            method=request_type,
            data=parameters,
            headers=headers,
        )
        records = res['data']
    except MonzoPermissionsError:
        records = {'error': 'Permissions error'}
    return records



"""Entry point for Monzo Viewer module."""
from monzo_viewer.app import app


def start(local_only: bool = True, port: int = 5000, debug: bool = False):
    """
    Entry point for Monzo Viewer.

    Args:
        local_only: Only listen to local interface if True otherwise listen to all
        port: Port to list on
        debug: Start in debugging mode if True
    """
    host = '127.0.0.1' if local_only else '0.0.0.0'
    app.run(host=host, port=port, debug=debug)


if __name__ == '__main__':
    start()
